#  ICE Revision: $Id: PlotRunner.py 8227 2007-11-23 14:02:58Z bgschaid $ 
"""
Class that implements pyFoamPlotRunner
"""

from PyFoamApplication import PyFoamApplication

from PyFoam.FoamInformation import changeFoamVersion

from PyFoam.Execution.GnuplotRunner import GnuplotRunner

from PyFoam.Execution.ParallelExecution import LAMMachine

from PyFoam.RunDictionary.SolutionDirectory import SolutionDirectory

from PyFoam.Error import warning

from os import path

class PlotRunner(PyFoamApplication):
    def __init__(self,args=None):
        description="""
        runs an OpenFoam solver needs the usual 3 arguments (<solver>
        <directory> <case>) and passes them on (plus additional arguments).
        Output is sent to stdout and a logfile inside the case directory
        (PyFoamSolver.logfile) Information about the residuals is output as
        graphs
        
        If the directory contains a file customRegexp this is automatically
        read and the regular expressions in it are displayed
        """
        PyFoamApplication.__init__(self,args=args,description=description)

    def addOptions(self):
        self.parser.add_option("--frequency",
                               type="float",
                               dest="frequency",
                               default=1.,
                               help="The frequency with which output should be generated (in seconds)")
        
        self.parser.add_option("--persist",
                               action="store_true",
                               dest="persist",
                               default=True,
                               help="Gnuplot windows stay after interrupt")
        
        self.parser.add_option("--non-persist",
                               action="store_false",
                               dest="persist",
                               help="Gnuplot windows close after interrupt")
        
        self.parser.add_option("--raise",
                               action="store_true",
                               dest="raiseit",
                               help="Raise the Gnuplot windows after every replot")
        
        self.parser.add_option("--no-default",
                               action="store_true",
                               default=False,
                               dest="nodefault",
                               help="Switch off the default plots (linear, continuity and bound)")
        
        self.parser.add_option("--no-linear",
                               action="store_false",
                               default=True,
                               dest="linear",
                               help="Don't plot the linear solver convergence")
        
        self.parser.add_option("--no-continuity",
                               action="store_false",
                               default=True,
                               dest="cont",
                               help="Don't plot the continuity info")
        
        self.parser.add_option("--no-bound",
                               action="store_false",
                               default=True,
                               dest="bound",
                               help="Don't plot the bounding of variables")
        
        self.parser.add_option("--with-iterations",
                               action="store_true",
                               default=False,
                               dest="iterations",
                               help="Plot the number of iterations of the linear solver")
        
        self.parser.add_option("--with-courant",
                               action="store_true",
                               default=False,
                               dest="courant",
                               help="Plot the courant-numbers of the flow")
        
        self.parser.add_option("--with-execution",
                               action="store_true",
                               default=False,
                               dest="execution",
                               help="Plot the execution time of each time-step")
        
        self.parser.add_option("--with-deltat",
                               action="store_true",
                               default=False,
                               dest="deltaT",
                               help="'Plot the timestep-size time-step")
        
        self.parser.add_option("--with-all",
                               action="store_true",
                               default=False,
                               dest="withAll",
                               help="Switch all possible plots on")
        
        self.parser.add_option("--custom-regexp",
                               action="append",
                               default=None,
                               dest="customRegex",
                               help="Add a custom regular expression to be plotted (can be used more than once)")
        
        self.parser.add_option("--write-files",
                               action="store_true",
                               default=False,
                               dest="writeFiles",
                               help="Writes the parsed data to files")
        
        self.parser.add_option("--regexp-file",
                               action="store",
                               default=None,
                               dest="regexpFile",
                               help="A file with regulare expressions that are treated like the expressions given with --custom-regexp")
        
        self.parser.add_option("--no-auto-customRegexp",
                               action="store_false",
                               default=True,
                               dest="autoCustom",
                               help="Do not automatically load the expressions from the file customRegexp")
        
        self.parser.add_option("--procnr",
                               type="int",
                               dest="procnr",
                               default=None,
                               help="The number of processors the run should be started on")
        
        self.parser.add_option("--machinefile",
                               dest="machinefile",
                               default=None,
                               help="The machinefile that specifies the parallel machine")
        
        self.parser.add_option("--clear-case",
                               action="store_true",
                               default=False,
                               dest="clearCase",
                               help="Clear all timesteps except for the first before running")
        
        self.parser.add_option("--steady-run",
                               action="store_true",
                               default=False,
                               dest="steady",
                               help="This is a steady run. Stop it after convergence")
        
        self.parser.add_option("--restart",
                               action="store_true",
                               default=False,
                               dest="restart",
                               help="Restart the simulation from the last time-step")
        
        self.parser.add_option("--progress",
                               action="store_true",
                               default=False,
                               dest="progress",
                               help="Only prints the progress of the simulation, but swallows all the other output")
        
        self.parser.add_option("--foamVersion",
                               dest="foamVersion",
                               default=None,
                               help="Change the OpenFOAM-version that is to be used")

        self.parser.add_option("--safe-until",
                               type="float",
                               dest="safeUntil",
                               default=None,
                               help="Sets lower under-relaxation and lower-order convection-schemes for the start of the simulation (only for steady-runs)")
        
        self.parser.add_option("--safe-relaxation-factor",
                               type="float",
                               dest="safeRelaxation",
                               default=0.5,
                               help="The factor by which the relaxation-factors should be scaled down (when calculating safe)")
        
        self.parser.add_option("--report-usage",
                               action="store_true",
                               default=False,
                               dest="reportUsage",
                               help="After the execution the maximum memory usage is printed to the screen")

        
    def addFileRegexps(self,fName):
        """Adds the lines from a file to the custom regular expressions
        @param fName: The name of the file"""
        f=open(fName)

        for l in f.readlines():
            l=l.strip()
            if len(l)==0:
                continue
            if l[0]=='"' and l[-1]=='"':
                l=l[1:-1]
            if len(l)>0:
                if self.opts.customRegex==None:
                    self.opts.customRegex=[]
                self.opts.customRegex.append(l)
        f.close()

        
    def run(self):
        if self.opts.foamVersion!=None:
            changeFoamVersion(self.opts.foamVersion)

        if self.opts.nodefault:
            self.opts.linear=False
            self.opts.cont=False
            self.opts.bound=False
            
        if self.opts.withAll:
            self.opts.linear=True
            self.opts.cont=True
            self.opts.bound=True
            self.opts.iterations=True
            self.opts.courant=True
            self.opts.execution=True
            self.opts.deltaT=True

        if self.opts.regexpFile!=None:
            self.addFileRegexps(self.opts.regexpFile)

        if self.opts.autoCustom:
            autoFile=path.join(self.parser.getArgs()[1],self.parser.getArgs()[2],"customRegexp")
            if path.exists(autoFile):
                print " Reading regular expressions from",autoFile
                self.addFileRegexps(autoFile)

        cName=self.parser.getArgs()[2]
        sol=SolutionDirectory(cName,archive=None)
        
        if self.opts.clearCase:
            print "Clearing out old timesteps ...."            
            sol.clearResults()

        lam=None
        if self.opts.procnr!=None or self.opts.machinefile!=None:
            lam=LAMMachine(machines=self.opts.machinefile,nr=self.opts.procnr)
        
        run=GnuplotRunner(argv=self.parser.getArgs(),smallestFreq=self.opts.frequency,persist=self.opts.persist,plotLinear=self.opts.linear,plotCont=self.opts.cont,plotBound=self.opts.bound,plotIterations=self.opts.iterations,plotCourant=self.opts.courant,plotExecution=self.opts.execution,plotDeltaT=self.opts.deltaT,customRegexp=self.opts.customRegex,writeFiles=self.opts.writeFiles,server=True,lam=lam,raiseit=self.opts.raiseit,steady=self.opts.steady,progress=self.opts.progress,restart=self.opts.restart)

        if self.opts.safeUntil:
            if not self.opts.steady:
                warning("This is an unsteady run. No safe settings set")
            else:
                warning("Adding Trigger and resetting to safer start-settings")
                trig=SafeTrigger(sol,self.opts.safeRelaxation)
                run.addTrigger(self.opts.safeUntil,trig.resetIt)
                run.addEndTrigger(trig.resetIt)
        
        run.start()

        if self.opts.reportUsage:
            print "\n  Used Memory: ",run.run.usedMemory(),"MB"

import re
from PyFoam.RunDictionary.ParsedParameterFile import ParsedParameterFile

class SafeTrigger:
    def __init__(self,sol,factor):
        self.solution=ParsedParameterFile(path.join(sol.systemDir(),"fvSolution"),backup=True)
        self.schemes=ParsedParameterFile(path.join(sol.systemDir(),"fvSchemes"),backup=True)

        self.fresh=True
        
        try:
            relax=self.solution["relaxationFactors"]
            for var in relax:
                relax[var]*=factor

            cExp=re.compile("div\((.+),(.+)\)")
            conv=self.schemes["divSchemes"]
            for nm in conv:
                if cExp.match(nm) or nm=="default":
                    conv[nm]="Gauss upwind"

            self.solution.writeFile()
            self.schemes.writeFile()
        except Exception,e:
            warning("Restoring defaults")
            self.solution.restore()
            self.schemes.restore()
            raise e
        
    def resetIt(self):
        if self.fresh:
            warning("Trigger called: Resetting fvSchemes and fvSolution")
            self.solution.restore()
            self.schemes.restore()
            self.fresh=False
